#!/usr/bin/python
#coding: utf-8

import shmht
import time

capacity = 300000

fd = shmht.open('/dev/shm/test.performance', capacity, 1)

begin_time = time.time()
for i in range(capacity):
    s = '%064d' % i
    shmht.setval(fd, s, s)
end_time = time.time()
print(capacity / (end_time - begin_time), 'iops @ set')

begin_timend_time = time.time()
for i in range(capacity):
    s = '%064d' % i
    if s != shmht.getval(fd, s):
        raise Exception(s)
end_time = time.time()
print(capacity / (end_time - begin_time), 'iops @ get')

shmht.close(fd)



#!/usr/bin/python
import os
from distutils.core import setup, Extension

shmht = Extension('shmht',
    sources = ['shmht.c', 'hashtable.c'],
    extra_compile_args = '-g3 -Wall -Werror -O3 -march=native'.split(),
)

setup(
    name            = 'pyshmht3',
    version         = '0.1.4',
    author          = 'muyakongali',
    author_email    = 'muyakongali@gmail.com',
    description     = 'provide sharing memory based hash table for python',
    license         = "BSD",
    keywords        = "python extension sharing memory based hash table",
    url             = "http://github.com/felix021/pyshmht",
    ext_modules     = [shmht],
    packages        = ["pyshmht3"]
)



#!/usr/bin/python
#coding: utf-8

import marshal
from  .HashTable import HashTable

_debug = False

class Cacher(object):
    """
    Cacher: wrap HashTable with serializer and write_back mechanism
        if you intend to modify the cache, call write_back() before the program exits

        notice:
            Cacher tries to simulate dict in most cases, mainly except for:
                (a) no __iter__, please use foreach instead
                (b) key should always be a str, where dict allows all hashable objects 
                (c) no comparation with other 'dict's
            When necessary, you can use .to_dict() to get a real dict object.
    """
    def __init__(self, name, capacity=0, force_init=False, serializer=marshal):
        """
        'name'          the path of the file to be 'mmap'ed
                        use MemCacher(name, ...) to add prefix '/dev/shm' automatically
        'capacity'      optional, if you want to connect to an existing shmht
        'serializer'    should contain loads/dumps (marshal, json, pickle, etc.)
        """

        self.ht = HashTable.HashTable(name, capacity, force_init, serializer)
        self.d = {}
        self.loads = serializer.loads
        self.dumps = serializer.dumps

    def __getitem__(self, key):
        d = self.d
        if key in d:
            val = d[key]
        else:
            val = self.loads(self.ht[key])
            d[key] = val
        return val

    def __setitem__(self, key, val):
        self.d[key] = val

    def __delitem__(self, key):
        if key in self.d:
            del self.d[key]
            try:
                del self.ht[key]
            except:
                pass
        else:
            del self.d[key]

    def __contains__(self, key): #notice: key will be cached here
        return self.get(key) != None

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except:
            return default

    def update(self, dic):
        self.d.update(dic)

    def foreach(self, callback):
        self.write_back()
        return self.ht.foreach(callback, unserialize=True)

    def to_dict(self):
        self.write_back()
        return self.ht.to_dict(unserialize=True)

    def write_back(self):
        self.ht.update(self.d, serialize=True)

    def close(self):
        if self.d:
            global _debug
            if not _debug:
                self.write_back() #commented out for testing
            del self.d
            self.d = None
        if self.ht:
            self.ht.close()
            self.ht = None

    def __del__(self):
        """
        don't rely on this, please call write_back() manually if necessary.
        """
        self.close()


def MemCacher(name, capacity=0, force_init=False, serializer=marshal):
    """
    Add an prefix '/dev/shm/' to `name`, so that the file is saved only in memory
    For more information, see `help(Cacher)`
    """
    name = '/dev/shm/' + name
    return Cacher(name, capacity, force_init, serializer)

if __name__ == "__main__":
    #test cases
    ht = MemCacher('test.Cacher', 1024, True)
    print('fd:', ht.ht.fd)

    #set
    ht['a'] = '1'
    ht['b'] = 2
    c = {'hello': 'world'}
    ht['c'] = c

    #get
    print(ht['b'] == 2)
    print(ht['c'] == c)
    print(ht.get('c') == c)
    print(ht.get('d') == None)
    try:
        ht['d']
        print(False)
    except:
        print(True)

    #contains
    print(('c' in ht) == True)
    print(('d' in ht) == False)

    #del
    del ht['c']
    print(ht.get('c') == None)
    try:
        del ht['d']
        print('del:', False)
    except:
        print(True)

    #update & to_dict & foreach
    dumps = marshal.dumps
    ht['c'] = c
    print(ht.to_dict() == {'a': '1', 'b': 2, 'c': c})

    def cb(key, value):
        global s
        s += key + str(value)

    s = ''
    ht.foreach(cb)
    print(s == 'a1b2c' + str(c))

    ht.update({'a': 'x', 'b': 1000})

    s = ''
    ht.foreach(cb)
    print(s == 'axb1000c' + str(c))

    print(ht.to_dict() == {'a': 'x', 'b': 1000, 'c': c})

    #close
    ht.close()
    try:
        ht['a']
        print(False)
    except:
        print(True)

    #write_back
    ht = MemCacher('test.Cacher', 1024, True)
    print('fd:', ht.ht.fd)
    ht['a'] = 1
    ht.write_back()
    ht['b'] = 2

    _debug = True
    ht.close() #write_back() is called in close() when not debugging

    ht = MemCacher('test.Cacher', 1024, False)
    print('fd:', ht.ht.fd)
    print(ht['a'] == 1)
    try:
        print(ht['b'])
        print(False)
    except:
        print(True)
    ht.close()

    #simple performance test
    import time

    capacity = 300000

    ht = MemCacher('test.Cacher', capacity, force_init=True)
    begin_time = time.time()
    for i in range(capacity):
        s = '%064d' % i
        ht[s] = s
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ set / no write_back ')

    ht.write_back()
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ set / after write_back ')

    ht.d = {}
    begin_time = time.time()
    for i in range(capacity):
        s = '%064d' % i
        if s != ht[s]:
            raise Exception(s)
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ get / no cache ')

    begin_time = time.time()
    for i in range(capacity):
        s = '%064d' % i
        if s != ht[s]:
            raise Exception(s)
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ get / all cached ')

    ht.close()



#!/usr/bin/python
#coding: utf-8

import shmht
import marshal

#basic wrapper: open, close, get, set, remove, foreach
#extended wrapper: getobj, setobj, [], to_dict, update

class HashTable(object):
    """
    Basic wrapper for shmht. For more information, see 'help(Cacher)'
    """
    def __init__(self, name, capacity=0, force_init=False, serializer=marshal):
        force_init = 1 if force_init else 0
        self.fd = shmht.open(name, capacity, force_init)
        self.loads = serializer.loads
        self.dumps = serializer.dumps

    def close(self):
        shmht.close(self.fd)

    def get(self, key, default=None):
        val = shmht.getval(self.fd, key)
        if val == None:
            return default
        return val

    def set(self, key, value):
        return shmht.setval(self.fd, key, value)

    def remove(self, key):
        return shmht.remove(self.fd, key)

    def foreach(self, callback, unserialize=False):
        if not unserialize:
            cb = callback
        else:
            loads = self.loads
            def mcb(key, value):
                return callback(key, loads(value))
            cb = mcb
        return shmht.foreach(self.fd, cb)

    def getobj(self, key, default=None):
        val = self.get(key, default)
        if val == default:
            return default
        return self.loads(val)

    def setobj(self, key, val):
        val = self.dumps(val)
        return self.set(key, val)

    def __getitem__(self, key):
        val = shmht.getval(self.fd, key)
        if val == None:
            raise KeyError(key)
        return val

    def __setitem__(self, key, value):
        return shmht.setval(self.fd, key, value)

    def __delitem__(self, key):
        if False == shmht.remove(self.fd, key):
            raise KeyError(key)

    def __contains__(self, key):
        return shmht.getval(self.fd, key) != None

    def to_dict(self, unserialize=False):
        d = {}
        def insert(k,v):
            d[k.decode()] = v
        self.foreach(insert, unserialize)
        return d

    def update(self, d, serialize=False):
        dumps = self.dumps
        if serialize:
            for k in d:
                self[k] = dumps(d[k])
        else:
            for k in d:
                self[k] = d[k]

if __name__ == "__main__":
    loads = marshal.loads
    dumps = marshal.dumps
    #test cases
    ht = HashTable('/dev/shm/test.HashTable', 1024, 1)

    #set
    ht['a'] = '1'
    ht.set('b', '2')
    c = {'hello': 'world'}
    ht.setobj('c', c)

    #get
    assert ht['b'] == b'2'
    assert ht['c'] == marshal.dumps(c)
    assert ht.getobj('c') == c
    assert ht.get('d') is None
    try:
        ht['d']
        assert False
    except: pass

    #contains
    assert 'c' in ht
    assert 'd' not in ht

    #del
    del ht['c']
    assert ht.get('c') is None
    try:
        del ht['d']
        assert False
    except: pass

    #update & to_dict & foreach
    ht.setobj('c', c)
    print(repr(ht.to_dict()))
    assert ht.to_dict() == {'a': b'1', 'b': b'2', 'c': dumps(c)}

    s = b''
    def cb(key, value):
        global s
        s += key + value
        print("key: %r, value: %r, s: %r" % (key,value, s))
    ht.foreach(cb)
    print(s)
    assert s == b'a1b2c' + dumps(c)

    ht.update({'a': 1, 'b': 2, 'c': c}, serialize=True)
    s = ''
    def cb2(key, value):
        global s
        s += key.decode() + str(value)
        print("key: %r, value: %r, s: %r" % (key,value, s))
    ht.foreach(cb2, unserialize=True)
    assert s == 'a1b2c' + str(c)

    assert ht.to_dict() == {'a':dumps(1), 'b':dumps(2), 'c':dumps(c)}
    assert ht.to_dict(unserialize=True) == {'a': 1, 'b': 2, 'c': c}

    #close
    ht.close()
    try:
        ht['a']
        assert False
    except: pass

    #simple performance test
    import time

    capacity = 300000

    #write_through
    ht = HashTable('/dev/shm/test.HashTable', capacity, True)

    begin_time = time.time()
    for i in range(capacity):
        s = str(i)
        ht[s] = s
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ set')

    begin_timend_time = time.time()
    for i in range(capacity):
        s = str(i)
        assert s == ht[s].decode()
    end_time = time.time()
    print(capacity / (end_time - begin_time), 'iops @ get')

    ht.close()
